import carinata
import os
import sys
from optparse import make_option

from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    args = '<app>'
    option_list = BaseCommand.option_list + (
        make_option('-g', '--generate', action="store_true", default=False,
                    help="Only generate test files, do not run them"
                    " (False by default, so tests will run if this"
                    " argument is not given)"),
        make_option("-f", "--force", action="store_true", default=False,
                    help="Generate test files regardless of whether"
                    " original files have changed or not (False by"
                    " default, so only changed tests are generated)"),
        make_option("-c", "--clean", action="store_true", default=False,
                    help="Clean up files instead of generating them"),

    )

    def handle(self, *apps, **options):
        if not apps:
            sys.stderr.write("""\
Please specify at least one app to run specs on.
""")
            sys.exit(1)

        generate = options.pop('generate')
        force = options.pop('force')
        clean = options.pop('clean')
        for app in apps:
            directories = [os.path.join(app, "spec")]
            carinata.main(directories, os.path.join(app, "tests"),
                          generate=True, force=force, clean=clean)
        if not generate and not clean:
            call_command('test', app)



#!/usr/bin/env python
# coding: utf-8
"""
carinata: a (rough-scaled) python spec generator.
Copyright (C) 2013  Scott McGinness <mcginness.s@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import os
import re
import sys
import tempfile
import unittest

from . import utils
from .block import Block
from .creator import Creator


MATCH = re.compile(r'''
    ^(?P<indent>\s*)   # whitespace indent
    (?P<name>describe|context|before|after|let|it)  # block name
    \s"                # space, then open quote
    (?P<words>[^"]*?)  # words of description
    ":                 # close quote, then end with colon
    \s?(?P<rest>.*)$   # code at end
''', re.VERBOSE)


class TestGenerator(object):
    def __init__(self, filepath, stream=sys.stdout):
        """Read the contents of a spec file and setup the generator"""
        with open(filepath) as file_to_read:
            contents = file_to_read.read()
        self.filehash = utils.get_hash_from_contents(contents)
        self.filepath = os.path.abspath(filepath)
        self.creator = Creator(stream)
        self.lines = contents.split("\n")
        self.blocks = [Block("", Block.test, "", 0)]
        self.deferred_its = []

    def process(self):
        """Do the main processing of the spec file"""
        self.creator.filehash(self.filehash)
        self.creator.notice(self.filepath)
        for lineno, line in enumerate(self.lines):
            line_match = MATCH.match(line)
            if line_match:
                self.process_line_match(lineno+1, line_match)
            elif not line or line.isspace():
                continue
            else:
                self.process_code(lineno+1, line)

    def process_line_match(self, lineno, line_match):
        """If the line matched a block, process that block"""
        indent, name, words, rest = line_match.groups()

        if name != Block.it and self.deferred_its:
            self.process_its()

        block = Block(indent, name, words, lineno, rest)

        if len(self.blocks) != 1 and block.indent <= self.blocks[-1].indent:
            self.blocks = [b for b in self.blocks
                           if b.is_applicable(block.indent)]
        self.blocks.append(block)

        if block.name == Block.it:
            self.defer_it()

    def process_code(self, lineno, line):
        """Write code lines into stream or append to block"""
        if len(self.blocks) == 1:
            # At the top level, so write immediately
            self.creator.line(line, lineno)
        else:
            # Inside a block, so defer it
            self.blocks[-1].code.append((lineno, line))

    def defer_it(self):
        """Put an ‘it’ block on a list to be bundled into a test class"""
        self.deferred_its.append(self.blocks[-1])

    def process_its(self):
        """Process the ‘it’ blocks given the context in the spec file"""
        structures, setups, teardowns = self.split_block_types()

        self.creator.klass(structures)
        for block in structures:
            self.creator.code(block, class_level=True)
        for setup in setups:
            self.creator.part_set_up(setup)
        for teardown in teardowns:
            self.creator.part_tear_down(teardown)
        if setups:
            self.creator.full_set_up(setups)
        if teardowns:
            self.creator.full_tear_down(teardowns)

        for it in self.deferred_its:
            self.creator.test(it)
        self.creator.line()
        self.deferred_its = []

    def split_block_types(self):
        """Split the list of blocks into structural and setup code"""
        structures, setups, teardowns = [], [], []
        for block in self.blocks:
            if block.name in [Block.describe, Block.context]:
                structures.append(block)
            elif block.name in [Block.before, Block.let]:
                setups.append(block)
            elif block.name == Block.after:
                teardowns.append(block)
        return structures, setups, teardowns


class SuiteGenerator(object):
    """Generate python test files or a unittest.TestSuite"""
    SUFFIX = ".carinata"
    TEMPDIR = os.path.join(tempfile.gettempdir(), "carinata")

    def __init__(self, directories, output_dir=None, force_generation=False, clean=False):
        self.directories = directories
        self.output_dir = output_dir
        self.force_generation = force_generation
        self.clean = clean

    def spec_files(self):
        """Get a list of paths to spec files in directories"""
        for directory in self.directories:
            for root, _, filenames in os.walk(directory):
                for filename in filenames:
                    if os.path.splitext(filename)[1] == self.SUFFIX:
                        yield directory, os.path.join(root, filename)

    def create_test_files(self):
        """Create python test files from the spec files"""
        filepaths = []
        for indir, infile in self.spec_files():
            if self.clean:
                outdir, outfile = self._get_output(indir, infile)
                os.remove(outfile)
                continue
            try:
                with self.output_file(indir, infile) as outfile:
                    test = TestGenerator(infile, outfile)
                    test.process()
                path = outfile.name
            except utils.FileHashMatch as hash_match:
                path = hash_match.filename
            finally:
                filepaths.append(path)

        return filepaths

    def create_test_suite(self):
        """Create a unittest suite from the spec files"""
        suite, loader = unittest.TestSuite(), unittest.TestLoader()
        filepaths = self.create_test_files()
        for filepath in filepaths:
            test = utils.create_module_from_file(filepath)
            suite.addTest(loader.loadTestsFromModule(test))
        return suite

    def _get_output(self, indir, infile):
        outdir = self.output_dir if self.output_dir else self.TEMPDIR

        outfile = infile.replace(self.SUFFIX, ".py")
        outfile = os.path.join(outdir,
                               os.path.relpath(outfile, indir))
        outdir = os.path.dirname(outfile)

        return outdir, outfile

    def output_file(self, indir, infile):
        """Get a file-like object for output.

        If an output_dir was given, put the file in there. Otherwise, put it
        in the temporary directory.
        """
        outdir, outfile = self._get_output(indir, infile)
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        # Check the hash of the output file and raise exception if it matches
        # that of the input file
        if not self.force_generation:
            utils.check_file_hash(infile, outfile)

        return open(outfile, 'w')


def main(directories, output_dir, generate, force, clean):
    """Generate and run spec files.

    Collect spec files from directories and process them into a test suite.
    If output_dir is given, put the test files into it, preserving directory
    structure from each parent directory. If generate is given, only generate
    the files, otherwise run with the usual unittest text runner.
    """
    generator = SuiteGenerator(directories, output_dir, force, clean)

    if generate:
        generator.create_test_files()
    else:
        suite = generator.create_test_suite()
        if not clean:
            unittest.TextTestRunner().run(suite)


def parse_args():
    """Define and parse command line arguments"""
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("directories", action="append",
                        help="The list of directories in which to search for"
                        " spec files (always recursive)")

    parser.add_argument("-o", "--output-dir", dest="output_dir",
                        help="The directory in which to create output test"
                        " files (Created in a temporary directory if this"
                        " argument is not given)")

    parser.add_argument("-g", "--generate", action="store_true", default=False,
                        help="Only generate test files, do not run them"
                        " (False by default, so tests will run if this"
                        " argument is not given)")

    parser.add_argument("-f", "--force", action="store_true", default=False,
                        help="Generate test files regardless of whether"
                        " original files have changed or not (False by"
                        " default, so only changed tests are generated)")

    parser.add_argument("-c", "--clean", action="store_true", default=False,
                        help="Clean up files instead of generating them")

    return parser.parse_args()


def main_cmdline():
    """Run carinata as main package, taking arguments from sys.argv"""
    args = parse_args()
    main(args.directories, args.output_dir, args.generate, args.force,
         args.clean)


if __name__ == '__main__':
    main_cmdline()



# coding: utf-8
"""Create a test class from blocks"""
import re

from .utils import camelify, snakify
from .block import Block

_4 = " " * 4
_8 = _4 * 2

LSTRIP = re.compile(r'[^\s]')


class Creator(object):
    _filehash = "# sha1: {0}\n"
    _notice = """\
# This file was auto-generated by carinata
# It may be overwritten at any time, so please refer to the original:
# {0}
#
"""

    _klass = "class Test{0}(TestCase):\n"
    _part_set_up = _4 + "def _set_up_{0}(self):\n"
    _full_set_up = _4 + "def setUp(self):\n"
    _part_tear_down = _4 + "def _tear_down_{0}(self):\n"
    _full_tear_down = _4 + "def tearDown(self):\n"
    _call_set_up = _8 + "self._set_up_{0}()\n"
    _call_tear_down = _8 + "self._tear_down_{0}()\n"
    _assign = _8 + "self.{0} = self._set_up_{0}()\n"
    _test = _4 + "def test_{0}(self):\n"
    _code = "{0}{1}  # L:{2}\n"

    def __init__(self, stream):
        """Write each part of a test class into a stream from blocks.

        For each method, there is a corresponding format string (called _method)
        which should describe what is written.
        """
        self.stream = stream

    def filehash(self, hexdigest):
        self.stream.write(self._filehash.format(hexdigest))

    def notice(self, filepath):
        self.stream.write(self._notice.format(filepath))

    def klass(self, blocks):
        """A class definition line, with name based on names of blocks"""
        name = "".join(camelify(block.words) for block in blocks)
        self.stream.write(self._klass.format(name))

    def part_set_up(self, block):
        """Write a partial _set_up_*() defintion with body"""
        self.stream.write(self._part_set_up.format(block.words))
        self.code(block)

    def part_tear_down(self, block):
        """Write a partial _tear_down_*() defintion with body"""
        self.stream.write(self._part_tear_down.format(block.words))
        self.code(block)

    def full_set_up(self, blocks):
        """Write the setUp() definition with body"""
        self.stream.write(self._full_set_up)
        for block in blocks:
            if block.name == Block.before:
                self.call(block)
            elif block.name == Block.let:
                self.assign(block)
        self.line()

    def full_tear_down(self, blocks):
        """Write the tearDown() definition with body"""
        self.stream.write(self._full_tear_down)
        for block in blocks:
            self.call(block)
        self.line()

    def call(self, block):
        """Write a call to a partial _set_up_*() method"""
        if block.name == Block.before:
            call = self._call_set_up
        elif block.name == Block.after:
            call = self._call_tear_down
        self.stream.write(call.format(block.words))

    def assign(self, block):
        """Write a call to a _set_up_*(), and assign to self.*"""
        self.stream.write(self._assign.format(block.words))

    def test(self, block):
        """Write a test_*() method with body"""
        name = snakify(block.words)
        self.stream.write(self._test.format(name))
        self.code(block)

    def code(self, block, class_level=False):
        """Write the code contained in block, dedenting where necessary"""
        if not block.code:
            return
        start = LSTRIP.search(block.code[0][1]).start()
        indent = _4 if class_level else _8
        fmt = lambda l, n: self._code.format(indent, l[start:], n)
        lines = [fmt(line, lineno) for (lineno, line) in block.code]
        self.stream.writelines(lines)
        self.line()

    def line(self, content="", suffix=None):
        """Write a new line, with optional content"""
        if suffix is not None:
            line = self._code.format("", content, suffix)
        else:
            line = content + "\n"
        self.stream.write(line)




# coding: utf-8
"""String utilities for creating unittest files from spec files"""
import hashlib
import os
import sys
import uuid


VALID_IDENTIFIER = ("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
                    "12345678990_")


class FileHashMatch(Exception):
    """The persisted file has a hash, and it matched the current file hash"""
    fmt = "File {0} matched old file with hash {1}"

    def __init__(self, filename, hash):
        self.filename = filename
        self.hash = hash
        self.message = self.fmt.format(filename, hash)


def _camel_safe(name):
    """Remove all non-identifier chars and underscores from name"""
    return identifier_safe(name).replace("_", "")


def identifier_safe(name):
    """Remove all non-identifier chars from name"""
    return "".join(char for char in name if char in VALID_IDENTIFIER)


def camelify(words):
    """Return a camel-cased version of words"""
    return "".join(_camel_safe(w) for w in words.title() if not w.isspace())


def snakify(words):
    """Return a snake-cased version of words"""
    return "_".join(identifier_safe(w) for w in words.lower().split())


def create_module_from_file(filepath):
    """Execute string as if it were a module, and return that module"""
    name = os.path.splitext(os.path.basename(filepath))[0]
    directory = os.path.dirname(filepath)
    if directory not in sys.path:
        sys.path.insert(0, directory)
    return __import__(name)


def uuid_hex(length=6):
    return "_x" + uuid.uuid4().hex[:length]


def get_hash_from_contents(contents):
    return hashlib.sha1(contents).hexdigest()


def get_hash_from_filename(filename, block=2**12):
    sha1 = hashlib.sha1()
    with open(filename) as f:
        while True:
            data = f.read(block)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def get_hash_from_first_line(filename):
    with open(filename, 'r') as f:
        hash_line = f.readline()
    try:
        return hash_line.split(':')[1].strip()
    except IndexError:
        return ""


def check_file_hash(input_filename, output_filename):
    current_hash = get_hash_from_filename(input_filename)
    if os.path.exists(output_filename):
        old_hash = get_hash_from_first_line(output_filename)
        if current_hash == old_hash:
            raise FileHashMatch(output_filename, current_hash)



# coding: utf-8
"""A simple block class to wrap spec blocks"""
from . import utils


class Block(object):
    """Represent a block structure by indent, name, description and code"""
    test = 'test'
    describe = 'describe'
    context = 'context'
    before = 'before'
    after = 'after'
    let = 'let'
    it = 'it'

    valid_children = {
        test: [describe],
        describe: [describe, context, before, after, let, it],
        context: [context, before, after, let, it],
        before: [],
        after: [],
        let: [],
        it: [],
    }

    def __init__(self, indent, name, words, lineno, rest=None):
        self.indent = len(indent)
        self.name = name
        self.words = words
        self.lineno = lineno
        self.code = []
        if rest:
            if self.name == self.let and not rest.startswith('return'):
                rest = "return (%s)" % rest
            self.code.append((lineno, rest))
        if self.name in [self.before, self.after]:
            self.words += utils.uuid_hex()

    def __repr__(self):
        return "<%s: %s>" % (self.name, self.words)

    def is_applicable(self, indent):
        """Determine whether this block applies at the given indent"""
        if self.name == self.test:
            return True
        elif self.indent > indent:
            return False
        elif self.indent == indent:
            return self.name not in [self.describe, self.context, self.it]
        return True



#!/usr/bin/env python
import os
from distutils.core import setup

def get_packages():
    packages = []
    for root, _, filenames in os.walk('carinata'):
        if '__init__.py' in filenames:
            packages.append(".".join(os.path.split(root)).strip("."))
    return packages

_long_description = """\
Carinata is a command line tool which transforms spec files into unittest cases.
It tries to be a bit like RSpec, but for python. It includes a management
command for Django.

Spec files contain blocks called ``describe``, ``context``, ``before``, ``after``,
``let`` and ``it``, which in turn contain pure python. Carinata uses these blocks
to create a ``TestCase`` corresponding to each ``it`` block, with the setup from
``before`` and ``let`` and the teardown from ``after``.

See the `project homepage`_ on GitHub for more information, but here is an example::

    describe "My Awesome class":
        context "with the number 42":
            let "awesome": Awesome(42)

            it "jumps for joy":
                assert self.awesome.jumps_for_joy()

        context "with a string":
            let "awesome": Awesome('wow!')

            it "says it":
                assert self.awesome.say() == "Awesome says 'wow!'"

.. _project homepage: https://github.com/scottmcginness/carinata
"""


setup(name='carinata',
      version='0.10.3',
      description='A rough-scaled python spec generator',
      long_description=_long_description,
      author='Scott McGinness',
      author_email='mcginness.s@gmail.com',
      url='https://github.com/scottmcginness/carinata',
      download_url='https://github.com/scottmcginness/carinata/archive/master.zip',
      packages=get_packages(),
      scripts=['scripts/carinata'],
      keywords=['testing', 'test', 'rspec'],
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: Console",
          "Framework :: Django",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Natural Language :: English",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Code Generators",
          "Topic :: Software Development :: Quality Assurance",
          "Topic :: Software Development :: Testing",
      ]
)

